'''
Polynomial Approximation
'''
from smolyak.applications.polynomials.polynomial_approximation import PolynomialApproximator
from smolyak.applications.polynomials.probability_distributions import ProbabilityDistribution
